###############################################################################
#
#   Copyright (c) 2006 Zenoss, Inc. all rights reserved.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License.
#
###############################################################################

#
# contained below are functions that are associated with installing or
# configuring zenoss.  these include things like initializing the
# mysql database or populating the zeo database.  you will not find
# functions here handle external dependency compilation or zenoss
# downloading (those are found in build-functions)
#
# Christopher Blunck
#

test -r shared-functions.sh && . shared-functions.sh

# return a list of pids and commands of zenoss related processes
get_running()
{
    FORMAT="pid,command"
    if [ "`uname`" = "SunOS" ]; then
        FORMAT="pid,comm"
    fi

    ps -Ao ${FORMAT} | grep python | egrep -v 'grep|buildbot' | egrep 'zeo|zope|zen' 
}

# use ps to see if zenoss processes are running.  fail if any are found
check_running()
{
    PS=`get_running`
    if [ "$PS" != "" ] ; then
        echo
        echo Looks like there are some stray zenoss processes running.
        echo ----------------------------------------------------------
        get_running
        echo ----------------------------------------------------------
        fail "Please kill these processes before running the install"
    fi
}

# kills off processes related to zenoss
kill_running()
{
    for pid in \
        `get_running | awk '{print $1}'`
    do
        kill $pid
    done
}

# attempt to create the events database and login to it.  fail if unsuccessful
create_database()
{
    if [ "${MYSQLHOST}" = "localhost" ]; then
        MYSQL_HOST_PERM="localhost"
    else
        MYSQL_HOST_PERM="%"
    fi

    mysql -h ${MYSQLHOST} -u ${MYSQLROOTUSER} --password="${MYSQLROOTPASSWD}" <<EOF
	    grant all on $MYSQLEVENTDB.* to '$MYSQLUSER'@'${MYSQL_HOST_PERM}' identified by '$MYSQLPASS';
	    grant super, file on *.* to '$MYSQLUSER'@'${MYSQL_HOST_PERM}'; 
	    flush privileges;
EOF

    # attempt to login to the database using the mysql user
    mysql -h ${MYSQLHOST} -u ${MYSQLUSER} -p"${MYSQLPASS}" < /dev/null >/dev/null || \
        fail mysql install, user or password is incorrect 
}

# populate the events database with some initial information
run_zeneventbuild()
{
    # Don't create the database if it exists
    if [ -f backup.tar ]; then
        echo "backup.tar exists.  skipping events database build"
        return
    fi
    
    echo "creating events database using zenevent build..." >> zenbuild.log
    $ZENHOME/bin/zeneventbuild $MYSQLHOST $MYSQLUSER $MYSQLPASS $MYSQLEVENTDB 2>>zenbuild.log || fail "failed to install event tables into MySQL"
}

# create a zope instance 
run_mkzopeinstance()
{
    USERNAME=${OS_USERNAME}
    id ${USERNAME} 2>&1 >/dev/null || USERNAME=$USER
    if [ ! -f backup.tar ] ; then
        $PYTHON $ZOPEHOME/bin/mkzopeinstance.py \
            --dir="$ZENHOME" \
            --user="admin:$ZOPEPASSWORD" || fail Unable to create zope instance.
        mv $ZENHOME/etc/zenoss.conf $ZENHOME/etc/zope.conf
        replace "effective-user zenoss" "effective-user $USERNAME" $ZENHOME/etc/zope.conf
    fi
}

# create a zeodb instance
run_mkzeoinstance()
{
    if [ ! -f backup.tar ] ; then
        $PYTHON $ZOPEHOME/bin/mkzeoinstance.py $ZENHOME || fail unable to create Zeo database instance.
    fi
}

# start up zeodb
start_zeodb()
{
    echo 'Starting Zope Object Database'
    $ZENHOME/bin/zeoctl start || fail unable to start Zeo object database
}

# shut down the zeodb
stop_zeodb()
{
    echo 'Stopping Zope Object Database'
    $ZENHOME/bin/zeoctl stop || fail unable to stop Zeo object database
}

# load objects into the zeodb
run_zenbuild()
{
    if [ ! -f backup.tar ] ; then
        echo Loading initial Zenoss objects into the Zeo database
        echo   '(this can take a few minutes)'
        $ZENHOME/bin/zenbuild -s $MYSQLHOST -u $MYSQLUSER -p "$MYSQLPASS" -d $MYSQLEVENTDB 2>>zenbuild.log || \
	          fail unable to create the initial Zenoss object database
    fi

    if [ "$MIGRATE" != "n" ]; then
        run_zenmigrate  >>zenbuild.log 2>&1 || fail Data migration failed
    fi

}

# migrate the zen database
run_zenmigrate()
{
    echo Migrating data
    $ZENHOME/bin/zenmigrate run
}

# starts the zope server
start_zope()
{
    echo 'Starting Zope Server'
    $ZENHOME/bin/zopectl start || fail unable to start Zope
}

# restart the snmpd process
restart_snmpd()
{
    /etc/init.d/snmpd restart
}

# restart the mysql process
restart_mysql()
{
    for script in \
        /etc/init.d/mysql \
        /etc/init.d/mysqld
    do
        if [ -x ${script} ]; then
            ${script} restart
            break
        fi
    done
}

# adds localhost as a device via the web UI
add_localhost_device()
{
    ZOPE_HOST=localhost.localdomain
    HOST_CLASS_PATH="/Server/Linux"
    URL="http://${ZOPE_HOST}:${ZOPE_LISTEN_PORT}/zport/dmd/DeviceLoader/loadDevice?deviceName=${ZOPE_HOST}&devicePath=${HOST_CLASS_PATH}"

    # add localhost as a device
    wget -O /dev/null \
        --user=${ZOPE_USERNAME} \
        --password=${ZOPE_PASSWORD} \
        ${URL} 

}

# updates the conf files in ${ZENHOME}/etc.  if the user already has a
# conf file then the example conf file will not be laid down on top of
# the existing conf file.  if the conf file already exists it is left
# in place.
update_conf_files()
{
    for EXAMPLE in \
        `ls ${ZENHOME}/etc/*.example`
    do
        CONF=`echo $EXAMPLE | cut -d. -f-2`
        
        if [ ! -s ${CONF} ]; then
            cp $EXAMPLE $CONF
        fi
    done
}
